# Pfilter/Config.pm
#
# library functions to handle pfilter configuration files

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# list of functions in this file:

# function to parse some address(s) from a config line
# sub parse_addresses( $vars_ref, $fields_ref, 
# 		         $error_lines_ref, $warning_lines_ref, 
# 		         $syntax_error_ref, $for, $where)

# function to parse some address(s) and/or interface(s) from a config line
# sub parse_addresses_interfaces( $vars_ref, $fields_ref, 
# 				$error_lines_ref, $warning_lines_ref, 
# 				$syntax_error_ref, $for, $where)

# function to parse one interface name from a config line
# sub parse_interface( $vars_ref, $fields_ref,
# 		       $error_lines_ref, $warning_lines_ref, 
# 		       $syntax_error_ref, $for, $where)

# function to parse one or more interface names from a config line
# sub parse_interfaces( $vars_ref, $fields_ref, 
# 		        $error_lines_ref, $warning_lines_ref, 
# 		        $syntax_error_ref, $for, $where)

# function to parse a keyword and some address(s) from a config line
# sub parse_keyword_addresses( $vars_ref, $fields_ref, $keyword, 
# 			       $error_lines_ref, $warning_lines_ref, 
# 			       $syntax_error_ref, $for, $where)

# function to parse a keyword and some interface(s) from a config line
# sub parse_keyword_interfaces( $vars_ref, $fields_ref, $keyword, 
# 			        $error_lines_ref, $warning_lines_ref, 
# 			        $syntax_error_ref, $for, $where)

# function to parse multiple keywords and address(s) from a config line
# sub parse_keywords_addresses( $vars_ref, $fields_ref, $keywords_wanted_ref, 
# 			        $error_lines_ref, $warning_lines_ref, 
# 			        $syntax_error_ref, $for, $where)

# function to parse multiple keywords and address(s) and/or interface(s)
# sub parse_keywords_addresses_interfaces( $vars_ref, $fields_ref,
#                                          $anticipated_keywords_ref, 
# 					   $address_keywords_wanted_ref, 
# 					   $interface_keywords_wanted_ref,
# 					   $error_lines_ref, $warning_lines_ref,
# 					   $syntax_error_ref, $for, $where)

# function to parse the pfilter configuration file that is in an array of lines.
# sub parse_pfilter_conf_lines( $vars_ref, $conf_lines_ref, 
# 			        $error_lines_ref, $warning_lines_ref )

# function to parse the the ports and/or services part of a config line
# sub parse_services_protocols_ports( $vars_ref, $fields_ref, 
# 				      $anticipated_keywords_ref,
# 				      $service_action, $error_lines_ref, 
# 				      $warning_lines_ref, $syntax_error_ref, 
# 				      $return_ok_with_unknown_input, 
# 				      $for, $where)

#=======================================================================

# function to parse some address(s) from a config line
#
# Passed parameters are:
#   reference to hash of useful flags/values
#   reference to array of words/fields left to be parsed
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be added to
#   reference to syntax error flag
#   string used to identify directive name in error lines
#   string used to identify directive location in error lines
# Returned parameters are:
#   array of returned addresses

sub parse_addresses
{
    my( $vars_ref, $fields_ref, $error_lines_ref,
	$warning_lines_ref, $syntax_error_ref, $for, $where) = @_;
    my @addresses = ();
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my $keywords_ref = $$vars_ref{keywords};

#    if ( $debug ) {
#	my $fields_string = join( ',', @$fields_ref);
#	print "$0: in parse_addresses fields=($fields_string)\n";
#    }

    # if no other arguments, error
    if ( ! @$fields_ref || $$keywords_ref{$$fields_ref[0]} ) {
	push @$error_lines_ref, "<$keyword> without addresses $for $where";
	$$syntax_error_ref = 1;
	return ();
    }

    # loop through arguments until we find another keyword
    while ( @$fields_ref && ! $$keywords_ref{$$fields_ref[0]} ) {
	my $address = shift @$fields_ref;
	my @ips = ();
	if ( valid_address ( $vars_ref, $address, \@ips ) ) {
	    push @addresses, $address;
	    
	} else {
	    push @$error_lines_ref,
	    "invalid address <$address> $for $where";
	    $$syntax_error_ref = 1;
	}
    }

    if ( $debug ) {
	my $string = join( ',', @addresses );
	print "$0: in parse_addresses returning ($string)\n";
    }
    return @addresses;
}

#=======================================================================

# function to parse some address(s) and/or interface(s) from a config line
#
# Passed parameters are:
#   reference to hash of useful flags/values
#   reference to array of words/fields left to be parsed
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be added to
#   reference to syntax error flag
#   string used to identify directive name in error lines
#   string used to identify directive location in error lines
# Returned parameters are:
#   reference to an array of returned addresses
#   reference to an array of returned interfaces

sub parse_addresses_interfaces
{
    my( $vars_ref, $fields_ref, $error_lines_ref,
	$warning_lines_ref, $syntax_error_ref, $for, $where) = @_;
    my @addresses = ();
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my @interfaces = ();
    my $keywords_ref = $$vars_ref{keywords};

#    if ( $debug ) {
#	my $fields_string = join( ',', @$fields_ref);
#	print "$0: in parse_addresses_interfaces fields=($fields_string)\n";
#    }

    # loop through arguments until we find another keyword
    while ( @$fields_ref && ! $$syntax_error_ref
	    && ! $$keywords_ref{$$fields_ref[0]} ) {
	my @ips = ();

	# if the current argument is "ALL" or "all", ...
	if ( $$fields_ref[0] eq "ALL" ||
	     $$fields_ref[0] eq "all" ) {
	    shift @$fields_ref;
	    push @interfaces, "ALL";
	    shift @$fields_ref;

	# or if the current argument is a valid address/range, ...
	} elsif ( valid_address ( $vars_ref, $$fields_ref[0], \@ips ) ) {
	    push @addresses, $$fields_ref[0];
	    shift @$fields_ref;

	# or if the current argument is an interface name, ...
	} elsif ( $$fields_ref[0] =~ /^[a-z]+[0-9]+$/ ) {
	    push @interfaces, $$fields_ref[0];
	    shift @$fields_ref;
	    
	} else {
	    push @$error_lines_ref,
	    "invalid address or interface <$$fields_ref[0]> $for $where";
	    $$syntax_error_ref = 1;
	}
    }

#    if ( $debug ) {
#	my $string_a = join( ',', @addresses );
#	my $string_i = join( ',', @interfaces );
#	print "$0: in parse_addresses_interfaces returning ($string_a, $string_i)\n";
#    }
    return ( \@addresses, \@interfaces );
}

#=======================================================================

# function to parse one interface name from a config line
#
# Passed input parameters are:
#   reference to hash of useful flags/values
#   reference to array of words/fields left to be parsed
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be possibly added to
#   reference to syntax error flag
#   string used to identify directive name in error lines
#   string used to identify directive location in error lines
# Returned parameters are:
#   parsed interface name(s) array
#     (if the argument is "ALL" or "all" every interface name is returned)

sub parse_interface
{
    my( $vars_ref, $fields_ref, $error_lines_ref, $warning_lines_ref,
	$syntax_error_ref, $for, $where) = @_;
    my $keywords_ref = $$vars_ref{keywords};
    my @parsed_interfaces;
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    
#    if ( $debug ) {
#	my $fields_string = join( ',', @$fields_ref);
#	print "$0: in parse_interface fields=($fields_string)\n";
#    }

    # parse an interface name if there is an argument, and there 
    # has been no previous syntax error, and if the next argument is
    # not a language keyword (which can't be used as an interface name)
    if ( @$fields_ref && ! $$syntax_error_ref
	 && ! $$keywords_ref{$$fields_ref[0]} ) {

	# if the current argument is "ALL" or "all", ...
	if ( $$fields_ref[0] eq "ALL" ||
	     $$fields_ref[0] eq "all" ) {
	    shift @$fields_ref;
	    @parsed_interfaces = ( "ALL" );
	    shift @$fields_ref;
	# if the current argument is an interface name, ...
	} elsif ( $$fields_ref[0] =~ /^[a-z]+[0-9]+$/ ) {
	    @parsed_interfaces = ( $$fields_ref[0] );
	    shift @$fields_ref;
	} else {
	    push @$error_lines_ref,
	    "$0: error - invalid interface name <$$fields_ref[0]> $for $where";
	    $$syntax_error_ref = 1;
	}

    }
#    if ( $debug ) {
#	my $string = join( ',', @$fields_ref);
#	print "$0: exiting from syntax_error=$$syntax_error_ref parse_interfaces fields=($string)\n";
#	$string = join( ',', @parsed_interfaces );
#	print "$0: in parse_interface parsed_interfaces = ($string)\n";
#    }

    return ( @parsed_interfaces );
}

#=======================================================================

# function to parse one or more interface names from a config line
#
# Passed input parameters are:
#   reference to hash of useful flags/values
#   reference to array of words/fields left to be parsed
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be possibly added to
#   reference to syntax error flag
#   string used to identify directive name in error lines
#   string used to identify directive location in error lines
# Returned parameters are:
#   parsed interfaces array

sub parse_interfaces
{
    my( $vars_ref, $fields_ref, $error_lines_ref, $warning_lines_ref,
	$syntax_error_ref, $for, $where) = @_;

    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my $keywords_ref = $$vars_ref{keywords};
    my @parsed_interfaces;

#    if ( $debug ) {
#	my $fields_string = join( ',', @$fields_ref);
#	print "$0: in parse_interfaces fields=($fields_string)\n";
#    }

    # loop through arguments, stopping when there are no more arguments,
    # or when there has been a syntax error, or when the next argument is
    # a language keyword (which can't be used as an interface name)
    while ( @$fields_ref && ! $$syntax_error_ref
	    && ! $$keywords_ref{$$fields_ref[0]} ) {

	# parse another interface
	my @interfaces = parse_interface
	    ($vars_ref, $fields_ref, $error_lines_ref,
	     $warning_lines_ref, $syntax_error_ref, $for, $where);
	# add it to the list of parsed interfaces
	push @parsed_interfaces, @interfaces
	    if !$$syntax_error_ref;
	# if there is an ALL in the list, remove any other interfaces
	# except the ALL since that implies any others
	if ( grep ( /^ALL$/, @parsed_interfaces ) ) {
	    my @nonalls = grep ( !/^ALL$/, @parsed_interfaces );
	    foreach my $nonall ( @nonalls ) {
		push @$warning_lines_ref,
		"removing interface $nonall because of ALL in interfaces list $for $where";
	    }
	    @parsed_interfaces = ( "ALL" )
		if ( @nonalls );
	}

    } # end of loop parsing interfaces
    if ( $debug ) {
	my $fields_string = join( ',', @$fields_ref);
	print "$0: exiting from parse_interfaces fields=($fields_string)\n";
	my $interfaces_string = join( ', ', @parsed_interfaces );
	print "$0:                               parsed_interfaces = ($interfaces_string)\n";
    }

    return ( @parsed_interfaces );
}

#=======================================================================

# function to parse a keyword and some address(s) from a config line
#
# Passed parameters are:
#   reference to hash of useful flags/values
#   reference to array of words/fields left to be parsed
#   keyword that triggers this
#   reference to system keywords hash
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be added to
#   reference to syntax error flag
#   string used to identify directive name in error lines
#   string used to identify directive location in error lines
# Returned parameters are:
#   array of returned addresses

sub parse_keyword_addresses
{
    my( $vars_ref, $fields_ref, $keyword, $error_lines_ref,
	$warning_lines_ref, $syntax_error_ref, $for, $where) = @_;
    my @addresses = ();
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my $keywords_ref = $$vars_ref{keywords};

    if ( $debug ) {
	my $fields_string = join( ',', @$fields_ref);
	print "$0: in parse_keyword_addresses keyword=$keyword fields=($fields_string)\n";
    }

    # if we have the correct keyword, ...
    if ( @$fields_ref && lc $$fields_ref[0] eq $keyword ) {
	shift @$fields_ref;

	# if no other arguments, error, otherwise, parse the addresses
	if ( ! @$fields_ref || $$keywords_ref{$$fields_ref[0]} ) {
	    push @$error_lines_ref, "<$keyword> without addresses $for $where";
	    $$syntax_error_ref = 1;
	    return ();
	} else {
	    @addresses = parse_addresses( $vars_ref,
					  $fields_ref, 
					  $error_lines_ref,
					  $warning_lines_ref,
					  $syntax_error_ref,
					  $for,
					  $where );
	}
    }

    if ( $debug ) {
	my $string = join( ',', @addresses );
	print "$0: in parse_keyword_addresses returning ($string)\n";
    }
    return @addresses;
}

#=======================================================================

# function to parse a keyword and some interface(s) from a config line
#
# Passed parameters are:
#   reference to hash of useful flags/values
#   reference to array of words/fields left to be parsed
#   keyword that triggers this
#   reference to system keywords hash
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be added to
#   reference to syntax error flag
#   string used to identify directive name in error lines
#   string used to identify directive location in error lines
# Returned parameters are:
#   array of returned addresses

sub parse_keyword_interfaces
{
    my( $vars_ref, $fields_ref, $keyword, $error_lines_ref,
	$warning_lines_ref, $syntax_error_ref, $for, $where) = @_;
    my @interfaces = ();
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my $keywords_ref = $$vars_ref{keywords};

    if ( $debug ) {
	my $fields_string = join( ',', @$fields_ref);
	print "$0: in parse_keyword_interfaces keyword=$keyword fields=($fields_string)\n";
    }

    # if we have the correct keyword, ...
    if ( @$fields_ref && lc $$fields_ref[0] eq $keyword ) {
	shift @$fields_ref;

	# if no other arguments, error, otherwise, parse the interfaces
	if ( ! @$fields_ref || $$keywords_ref{$$fields_ref[0]} ) {
	    push @$error_lines_ref, "<$keyword> without interfaces $for $where";
	    $$syntax_error_ref = 1;
	    return ();
	} else {
	    @interfaces = parse_interfaces( $vars_ref,
					    $fields_ref, 
					    $error_lines_ref,
					    $warning_lines_ref,
					    $syntax_error_ref,
					    $for,
					    $where );
	}
    }

    if ( $debug ) {
	my $string = join( ',', @interfaces );
	print "$0: in parse_keyword_interfaces returning ($string)\n";
    }
    return @interfaces;
}

#=======================================================================

# function to parse multiple keywords and address(s) from a config line
#
# Passed parameters are:
#   reference to hash of useful flags/values
#   reference to array of words/fields left to be parsed
#   reference to array of keywords that are allowable
#   reference to hash of system keywords
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be added to
#   reference to syntax error flag
#   string used to identify directive name in error lines
#   string used to identify directive location in error lines
# Returned parameters are:
#   hash with found keywords as keys and matching addresses lists as values

sub parse_keywords_addresses
{
    my( $vars_ref, $fields_ref, $keywords_wanted_ref, 
	$error_lines_ref, $warning_lines_ref, $syntax_error_ref,
	$for, $where) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my $keywords_ref = $$vars_ref{keywords};

    if ( $debug ) {
	my $keywords_string = join(',', @$keywords_wanted_ref);
	my $fields_string   = join(',', @$fields_ref);
	print "$0: in parse_keywords_addresses keywords=($keywords_string) fields=($fields_string)\n";
    }

    my %returned_values;
    if ( @$fields_ref ) {
	my $argument = lc $$fields_ref[0];
	while ( @$keywords_wanted_ref && @$fields_ref &&
		grep(/^$argument$/, @$keywords_wanted_ref) &&
		! $$syntax_error_ref ) {
	    my @addresses = parse_keyword_addresses
		( $vars_ref, $fields_ref, $argument,
		  $error_lines_ref, $warning_lines_ref, $syntax_error_ref,
		  $for, $where );
	    $returned_values{$argument} = \@addresses;
	    $argument = lc $$fields_ref[0];
	}
    }

    print_hash( "", "in parse_keywords_addresses returning", \%returned_values )
	if $debug;

    return \%returned_values;
}

#=======================================================================

# function to parse multiple keywords and address(s) and/or interface(s)
# from a config line
#
# Passed parameters are:
#   reference to hash of useful flags/values
#   reference to array of words/fields left to be parsed
#   reference to array of address keywords that are allowable
#   reference to array of interface keywords that are allowable
#   reference to hash of system keywords
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be added to
#   reference to syntax error flag
#   string used to identify directive name in error lines
#   string used to identify directive location in error lines
# Returned parameters are:
#   hash with found keywords as keys and matching addresses lists as values

sub parse_keywords_addresses_interfaces
{
    my( $vars_ref, $fields_ref, $anticipated_keywords_ref,
	$address_keywords_wanted_ref, $interface_keywords_wanted_ref,
	$error_lines_ref, $warning_lines_ref, $syntax_error_ref,
	$for, $where) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my $keywords_ref = $$vars_ref{keywords};

    if ( $debug ) {
	my $anticipated_keywords_string = join(',', @$anticipated_keywords_ref);
	my $address_keywords_string = join(',', @$address_keywords_wanted_ref);
	my $interface_keywords_string = join(',', @$interface_keywords_wanted_ref);
	my $fields_string   = join(',', @$fields_ref);
	print "$0: in parse_keywords_addresses_interfaces anticipated_keywords=($anticipated_keywords_string) address_keywords=($address_keywords_string) interface_keywords=($interface_keywords_string) fields=($fields_string)\n";
    }

    my %returned_values;
    if ( @$fields_ref ) {
	my $argument = lc $$fields_ref[0];
	while ( @$address_keywords_wanted_ref 
		&& 
		@$fields_ref
		&&
		( grep(/^$argument$/, @$address_keywords_wanted_ref) ||
		  grep(/^$argument$/, @$interface_keywords_wanted_ref) )
		&&
		! $$syntax_error_ref ) {
	    if ( grep(/^$argument$/, @$address_keywords_wanted_ref) ) {
		my @addresses = parse_keyword_addresses
		    ( $vars_ref, $fields_ref, $argument,
		      $error_lines_ref, $warning_lines_ref, $syntax_error_ref,
		      $for, $where );
		$returned_values{$argument} = \@addresses;
	    } else {
		my @interfaces = parse_keyword_interfaces
		    ( $vars_ref, $fields_ref, $argument,
		      $error_lines_ref, $warning_lines_ref, $syntax_error_ref,
		      $for, $where );
		$returned_values{$argument} = \@interfaces;
	    }
	    $argument = lc $$fields_ref[0] if @$fields_ref;
	}
    }

    print_hash( "", "in parse_keywords_addresses_interfaces returning", \%returned_values )
	if $debug;

    return \%returned_values;
}

#=======================================================================

# function to parse the pfilter configuration file that is in an array of lines.
#
# Passed parameters are:
#   reference to hash of flags/values to be updated for things like:
#     nowayit -> verb form of packet blocking ( "REJECT" or "DROP" )
#     nowayed -> adjective form of packet blocking ( "REJECTED" or "DROPPED" )
#   reference to array of configuration file lines without newline characters
#   reference to an array of error messages lines
#   reference to an array of warning message lines
# Returned parameters are:
#   reference to an array of references to directive/comment hash records,'
#   each directive or comment lines block record can contain:
#     directive_type => directive name string ("open" or whatever) if one there,
#                       or "comment_lines" for a block of comment lines
#     config_line_number => line number of directive
#     line => complete directive line string
#     comment_lines_num => number of comment lines to speed up calculations
#     comment_lines_ref => reference to an array of comment lines
#     tail_comment => comment string on end of directive line if there
#     destination_protocols_ports => reference to destination protocols/ports hash,
#                                    if appropriate for the directive,
#                                    and if there in the directive line,
#                                    organized as protocols beng hash key,
#                                    and value for each protocol hash being
#                                    reference to ports list
#     destination_services => reference to destination service(s) list, if appropriate
#                 for the directive, and if there in the directive line
#     interfaces => reference to incoming or outgoing interface name(s) list,
#                   if appropriate for the directive, and if there in the
#                   directive line
#     source_addresses => reference to source address(es) list, if
#                         appropriate for the directive, and if there in
#                         the directive line
#     destination_addresses => reference to destination address(es) list,
#                              if appropriate for the directive, and if
#                              there in the directive line
#     source_ports => reference to source port(s) list, if appropriate for
#                     the directive, and if there in the directive line
#     ???reference to protocols(s)/service(s)/port(s) list
#
# For example, after calling the function as follows:
#   ( $directives_ref, $errors_ref, $warnings_ref ) =
#     parse_pfilter_config( \%variables, \@conf_lines )
# if the fourth directive was an "open" services directive, some fields
# could be access like this:
#   $line             = $directives[3]{line};
#   $directive_string = $directives[3]{directive};
#   @source_addresses = $directives[3]{source_addresses};

sub parse_pfilter_conf_lines
{
    my( $vars_ref, $conf_lines_ref, $error_lines_ref,
	$warning_lines_ref ) = @_;
	
    my @comment_lines;
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my @directives;
    my $in_literal_block = 0;     # non-zero while in a literal block
    my $keywords_ref = $$vars_ref{keywords};
    my $literal_config_line;
    my @literal_lines;
    my $literal_line_num;
    # list of directives that have no arguments and require
    # no processing except for storing them
    my %noargs_directives = (
			     debug     => 1,
			     drop      => 1,
			     nomerge   => 1,
			     noverbose => 1,
			     reject    => 1,
			     verbose   => 1
			     );
    my $protocols_ref = $$vars_ref{protocols};

    # for each non-comment configuration file line
    foreach my $line_num ( 1 .. @$conf_lines_ref ) {
	my $config_file_line = $conf_lines_ref->[$line_num - 1];
	my @directive_error_lines = ();
	my @directive_warning_lines = ();

	# figure out if this line has a directive in it

	my $directive_text = $config_file_line;	# use different variable to protect list
	$directive_text =~ s/\s*#.*$//;		# get rid of comments
	$directive_text =~ s/^\s+//;		# get rid of leading white space
#	print "$0: parsing configuration line $line_num $config_file_line\<\n"
#	    if $debug;
	my $syntax_error = 0;
	my @fields = split(' ',$directive_text); # split the directive text into words
#	print "$0: configuration line has " . scalar @fields . " fields\n"
#	    if $debug;
	$fields[0] = lc $fields[0]; # directive word to lowercase for matching

	# if we're in the middle of a literal block,
	# and the directive is not the ENDLITERAL,
	# store this line in the current literal block
	if ( $in_literal_block &&
	     $fields[0] ne "endliteral" ) {
	    push @literal_lines, $config_file_line;
	    print "$0: literal_lines=@literal_lines\n" if $debug;
	    next;
	}

	# if no directive in the line, must be a comment line save it
	# on the end of the current comment lines block

	unless ( $directive_text ) {
	    push @comment_lines, $config_file_line;
	    next;
	}

	# there is a directive in the line, we have to parse and store it but
	# first we have to save any accumulated comment lines

	if ( @comment_lines ) {
	    my %comment_directive = (
		directive_type    => "comment_lines",
		comment_lines_ref => [@comment_lines],
		comment_lines_num => scalar @comment_lines
	    );
	    push @directives, \%comment_directive;
	    @comment_lines = ();
	}

	# assume we will find a good directive and help out the parsing storage

	my %directive = (
	    config_line_number => $line_num,
	    config_line        => $config_file_line,
	    directive_type     => $fields[0],
	    processed          => 0
	);
	my $tail_comment = $config_file_line;
	$tail_comment =~ s/^.*$directive_text//;
	$directive{tail_comment} = $tail_comment if $tail_comment;
	$directive{config_where} = 
	    "file ${$$vars_ref{constants}}{config_file_path} line $line_num";

	# help error messages code

	my $for   = "for $fields[0]";
	my $where = "on line $line_num in ${$$vars_ref{constants}}{config_file_path}";
	my $expecting = "";

	# check and save directives that have no arguments and that
	# require no further processing now except for storing them

	if ( $noargs_directives{$fields[0]} ) {
	    shift @fields;
	}

	# check and save valid ALIAS directives
	#
	# syntax: 
	#
	# ALIAS faked-address real-address \
	# [ protocols/ports/services ] [ [ FROM ] source[s] ]

	elsif ($fields[0] eq "alias") {
	    shift @fields;

	    # has to have at least two arguments
	    if (@fields < 2 ) {
		push @directive_error_lines, "not enough arguments $for $where";
		$syntax_error = 1;
	    } else {

		# add some fields to the directive block
		$directive{destination_protocols_ports} = {};
		$directive{destination_services}        = [];

		# get the faked pseudo address we need to create
		my $faked_address = shift @fields;
		my @ips = ();
		unless ( valid_address( $vars_ref, $faked_address, \@ips ) ) {
		    push @directive_error_lines,
			"invalid alias address <$faked_address> $for $where";
		    $syntax_error = 1;
		}
		$directive{faked_address} = $faked_address;

		# get the real hidden/protected address we want to send stuff to
		my $real_address = shift @fields;
		unless ( valid_internal_address($vars_ref, $real_address) ) {
		    push @directive_error_lines,
			"invalid aliased address <$real_address> $for $where";
		    $syntax_error = 1;
		}
		$directive{real_address} = $real_address;

		# parse destination ports/services to be opened
		my @anticipated_keywords = ( "from" );
		( $directive{destination_services},
		  $directive{destination_protocols_ports} ) =
		      parse_services_protocols_ports( $vars_ref,
						      \@fields, 
						      \@anticipated_keywords,
						      "open",
						      \@directive_error_lines,
						      \@directive_warning_lines,
						      \$syntax_error,
						      1,
						      $for,
						      $where);

		# since we still allow the old style syntax
		# (not requiring the "from" word before 
		# allowable from addresses), if the next
		# field is a valid internet address, output
		# a warning message and stick a from word in
		my @resolved_ips = ();
		if ( @fields &&
		     ! ( grep( /^$fields[0]$/, @anticipated_keywords ) ) &&
		     valid_address( $vars_ref, $fields[0], \@resolved_ips ) ) {
		    push @directive_warning_lines,
		    "Allowing source address(es) without the FROM keyword $for $where";
		    push @directive_warning_lines,
		    "This feature may be deprecated soon, please insert a FROM";
		    unshift @fields, "from";
		}

		# parse from address(es) if there
		my @source_addresses = 
		    parse_keyword_addresses( $vars_ref,
					     \@fields,
					     "from",
					     \@directive_error_lines,
					     \@directive_warning_lines,
					     \$syntax_error,
					     $for, 
					     $where);
		$directive{source_addresses} = \@source_addresses;
	    }
	}

	# check and save valid CLOSE/OPEN directives
	# (they both have the same syntax and options)
	#
	# syntax: 
	#
	# (CLOSE|OPEN) protocols/ports/services \
	# [ FROM ] source(s) ] | [ TO destination(s) ] | [ ON interface(s) ]

	elsif ( $fields[0] eq "close" ||
		$fields[0] eq "open" ) {
	    shift @fields;

	    # has to have at least one argument
	    if (@fields < 1) {
		push @directive_error_lines, "not enough arguments $for $where\n";
		$syntax_error = 1;
	    } else {

		# add some fields to the directive block
		my %destination_protocols_ports;
		$directive{destination_protocols_ports} =
		    \%destination_protocols_ports;
		$directive{source_addresses} = [];
		
		# parse destination ports/services to be closed/opened
		my @anticipated_keywords = ( "from", "to", "on" );
		( $directive{destination_services},
		  $directive{destination_protocols_ports} ) =
		      parse_services_protocols_ports( $vars_ref,
						      \@fields, 
						      \@anticipated_keywords,
						      $directive{directive_type},
						      \@directive_error_lines,
						      \@directive_warning_lines,
						      \$syntax_error,
						      1,
						      $for,
						      $where);
		
		# since we still allow the old style syntax
		# (not requiring the "from" word before 
		# allowable from addresses), if the next
		# field is a valid internet address, output
		# a warning message and stick a from word in
		my @resolved_ips = ();
		if ( @fields &&
		     ! ( grep( /^$fields[0]$/, @anticipated_keywords ) ) &&
		     valid_address( $vars_ref, $fields[0], \@resolved_ips ) ) {
		    push @directive_warning_lines,
		    "Allowing source address(es) without the FROM keyword $for $where";
		    push @directive_warning_lines,
		    "This feature may be deprecated soon, please insert a FROM";
		    unshift @fields, "from";
		}

		# parse from and/or destination addresss and/or interfaces if there
		@anticipated_keywords = ();
		my @address_keywords = ( "from", "to" );
		my @interface_keywords = ( "on" );
		my $results_hash_ref =
		    parse_keywords_addresses_interfaces( $vars_ref,
							 \@fields,
							 \@anticipated_keywords,
							 \@address_keywords,
							 \@interface_keywords,
							 \@directive_error_lines,
							 \@directive_warning_lines,
							 \$syntax_error,
							 $for,
							 $where);
		$directive{source_addresses} =
		    $$results_hash_ref{from} || [];
		$directive{destination_addresses} =
		    $$results_hash_ref{to}   || [];
		$directive{interfaces} =
		    $$results_hash_ref{on}   || [];
	    }
	}

	# handle ENDLITERAL directive

	elsif ( $fields[0] eq "endliteral" ) {
	    shift @fields;
	    # cannot be any arguments but we will still 
	    # just issue a warning if so
	    if (@fields) {
		push @directive_warning_lines,
		"endliteral requires no arguments $for $where";
		@fields = ();
	    }
	    # if we're not in the middle of a literal block
	    # this is an un-matched ENDLITERAL, declare an error
	    if ( ! $in_literal_block ) {
		push @directive_error_lines,
		"endliteral without previous literal $for $where";
		$syntax_error = 1;
	    }
	    # but if we are in a literal block, end it
	    else {
		delete $directive{config_line};
		delete $directive{config_line_number};
		delete $directive{config_where};
		$directive{directive_type} = "literal_block";
		$directive{endliteral_config_line} = $config_file_line;
		$directive{endliteral_config_line_number} = $line_num;
		$directive{literal_config_line} = $literal_config_line;
		$directive{literal_config_line_number} = $literal_config_line_number;
		$directive{literal_lines_ref} = [@literal_lines];
		$in_literal_block = 0;
	    }
	}

	# handle EXTIF/EXT INTERFACE(S)/EXTERNAL INTERFACE(S) directives

	elsif ( $fields[0] =~ /^extif$/i ||
	     ( @fields >= 2 && $fields[0] =~ /^ext(ernal)?$/i ) ||
	     ( @fields >= 3 && $fields[0] =~ /^ext(ernal)?$/i &&
	                       $fields[1] =~ /^interface(s)?$/i ) ) {
	    if ( $fields[0] eq "extif" ) {
		shift @fields;
	    } elsif ( $fields[1] =~ /^interface(s)?$/i ) {
		$for = "$for $fields[1]";
		$directive{directive_type} = "extif";
		shift @fields;
		shift @fields;
	    } else {
		$directive{directive_type} = "extif";
		shift @fields;
	    }
	    
	    # has to have at least one argument
	    if (@fields < 1 ) {
		push @directive_error_lines, "not enough arguments $for $where";
		$syntax_error = 1;
	    } else {
		
		# parse interface name(s)
		my @interfaces = parse_interfaces( $vars_ref,
						   \@fields,
						   \@directive_error_lines,
						   \@directive_warning_lines, 
						   \$syntax_error,
						   $for, 
						   $where);
		$directive{interfaces} = \@interfaces;
	    }
	}

	# handle FILTERED TRUSTED UNFILTERED UNTRUSTED directives
	#
	# syntax: 
	#
	# FILTERED [INTERFACE[S]] interface-name(s)
	# TRUSTED [INTERFACE[S]] interface-and-or-source-range(s) [ TO desination(s) ]
	# UNFILTERED [INTERFACE[S]] interface-name(s)
	# UNTRUSTED [INTERFACE[S]] interface-and-or-source-range(s) [ TO desination(s) ]

	elsif ( $fields[0] eq "filtered"    ||
	        $fields[0] eq "trusted"     ||
	        $fields[0] eq "unfiltered"  ||
	        $fields[0] eq "untrusted" ) {
	    shift @fields;

	    # get rid of possible INTERFACE(S) keyword
	    my $interface_next;
	    if ( $fields[0] =~ /^interface(s)?$/i ) {
		$for = "$for $fields[0]";
		shift @fields;
		$interface_next = 1;
	    } else {
		$interface_next = 0;
	    }

	    # has to have at least one argument
	    if (@fields < 1 ) {
		push @directive_error_lines, "not enough arguments $for $where";
		$syntax_error = 1;
	    } else {
		my $addresses_ref;
		my $interfaces_ref;
		my @interfaces = ();

		# if there was an INTERFACE(S) keyword, the next
		# thing parsed has to be an interface
		if ( $interface_next ) {
		    my @interface =
			parse_interface( $vars_ref,
					 \@fields,
					 \@directive_error_lines,
					 \@directive_warning_lines, 
					 \$syntax_error,
					 $for,
					 $where);
		    push @interfaces, @interface
			if ! $syntax_error;
		}

		# parse interface name(s) and/or address (range)(s)
		if ( ! $syntax_error ) {
		    ( $addresses_ref, $interfaces_ref )  =
			parse_addresses_interfaces( $vars_ref,
						    \@fields,
						    \@directive_error_lines,
						    \@directive_warning_lines, 
						    \$syntax_error,
						    $for,
						    $where);
		    push @interfaces, @$interfaces_ref
			if ! $syntax_error && @$interfaces_ref;
		}

		$directive{source_addresses} = $addresses_ref
		    if @$addresses_ref;
		$directive{interfaces} = \@interfaces;
		
		# if there were any parsed source address ranges,
		# try to parse TO destination(s)
		if ( @$addresses_ref ) {
		    my @keywords = ( "to" );
		    my $results_hash_ref =
			parse_keywords_addresses( $vars_ref,
						  \@fields,
						  \@keywords,
						  \@directive_error_lines,
						  \@directive_warning_lines,
						  \$syntax_error,
						  $for,
						  $where);
		    $directive{destination_addresses} =
			$$results_hash_ref{to}   || [];
		}
	    }
	}

	# check and save valid FORWARD directives
	#
	# syntax: 
	#
	# FORWARD protocols/ports/services ] [ FROM source-address(es) ] \
	# [ TO destination-address(es) ] [ ON interface(s) ] \
	# ONTO forwarded-destination-address \
	# [ forwarded-destination-protocols-ports-services ]

	elsif ($fields[0] eq "forward") {
	    shift @fields;

	    # add some fields to the directive block
	    $directive{forwarded_protocols_ports} = {};
	    $directive{forwarded_services}        = ();
	    
	    # parse ports/services to be forwarded
	    my @anticipated_keywords = 
		( "from", "to", "on", "onto" );
	    my( $forwarded_services_ref, $forwarded_protocols_ports_ref ) =
		parse_services_protocols_ports( $vars_ref,
						\@fields, 
						\@anticipated_keywords,
						"open",
						\@directive_error_lines, 
						\@directive_warning_lines,
						\$syntax_error,
						0,
						$for, 
						$where);
	    $directive{forwarded_services} = $forwarded_services_ref;
	    $directive{forwarded_protocols_ports} =
		$forwarded_protocols_ports_ref;
	    if ( ! @$forwarded_services_ref &&
		 ! %$forwarded_protocols_ports_ref ) {
		push @directive_error_lines,
		"missing forwarded services/protocols/ports $for $where";
		$syntax_error = 1;
	    }
	    
	    # parse from and/or to addresss if there
	    @anticipated_keywords = ( "onto" );
	    my @address_keywords = ( "from", "to" );
	    my @interface_keywords = ( "on" );
	    my $results_hash_ref =
		parse_keywords_addresses_interfaces( $vars_ref,
						     \@fields,
						     \@anticipated_keywords,
						     \@address_keywords,
						     \@interface_keywords,
						     \@directive_error_lines,
						     \@directive_warning_lines, 
						     \$syntax_error,
						     $for, 
						     $where);
	    $directive{source_addresses} =
		$$results_hash_ref{from} || [];
	    $directive{destination_addresses} =
		$$results_hash_ref{to} || [];
	    $directive{interfaces} =
		$$results_hash_ref{on}   || [];
	    
	    # now there has to an onto address
	    if ( ! @fields || $fields[0] ne "onto" ) {
		push @directive_error_lines, "no onto found $for $where";
		$syntax_error = 1;
	    } elsif ( scalar @fields < 2 ) {
		push @directive_error_lines, "missing address for onto $for $where";
		$syntax_error = 1;
	    } else {
		shift @fields;
		my $address = shift @fields;
		my @resolved_ips = ();
		if ( ! valid_address ( $vars_ref, $address, \@resolved_ips ) ) {
		    push @directive_error_lines,
		    "invalid onto address <$address> $for $where";
		    $syntax_error = 1;
		} else {
		    $directive{onto_address} = $address;

		    # and there can be an output port/icmp-type that
		    # packets get remapped to

		    my @onto_keywords = ( );
		    my( $onto_services_ref, $onto_protocols_ports_ref ) =
			parse_services_protocols_ports( $vars_ref,
							\@fields, 
							\@onto_keywords,
							"open",
							\@directive_error_lines, 
							\@directive_warning_lines,
							\$syntax_error,
							0,
							$for, 
							$where);

		    # if forwarded onto ruleset defined services were parsed, error
		    if ( @$onto_services_ref ) {
			$syntax_error = 1;
			foreach my $onto_service ( @$onto_services_ref ) {
			    push @directive_error_lines,
			    "cannot forward packets onto ruleset defined network service $onto_service $for $where";
			}
		    }

		    # if forwarded onto protocols/ports were parsed, check for problems
		    if ( %$onto_protocols_ports_ref ) {
			$directive{onto_protocols_ports} = $onto_protocols_ports_ref;

			# for every forwarded from protocol, ...
			foreach my $forwarded_protocol 
			    ( sort keys %$forwarded_protocols_ports_ref ) {

			    # every forwarded from protocol has to have a 
			    # matching forwarded onto same protocol
			    if ( ! exists $$onto_protocols_ports_ref{$forwarded_protocol} ) {
				$syntax_error = 1;
				push @directive_error_lines,
				"forwarded protocols include $forwarded_protocol which is not included in the forwarded onto protocols $for $where";
			    } else {

				# and the number of forwarded from ports for
				# that protocol has to match the number of 
				# forwarded onto ports for that protocol
				#print "forwarded_protocol=<$forwarded_protocol>\n";
				#print_hash("","forwarded_protocols_ports_ref",$forwarded_protocols_ports_ref);
				#print_hash("","onto_protocols_ports_ref",$onto_protocols_ports_ref);
				my $forwarded_ports_ref = 
				    $$forwarded_protocols_ports_ref{$forwarded_protocol};
				my $forwarded_ports_num = scalar keys %$forwarded_ports_ref;
				my $onto_ports_ref = 
				    $$onto_protocols_ports_ref{$forwarded_protocol};
				my $onto_ports_num = scalar keys %$onto_ports_ref;
				if ( $forwarded_ports_num != $onto_ports_num ) {
				    $syntax_error = 1;
				    push @directive_error_lines,
				    "forwarded protocol $forwarded_protocol has $forwarded_ports_num ports forwarded onto $onto_ports_num ports $for $where";
				}
                            } # end else if protocol in both forwarded and onto
			} # end foreach forwarded from protocol

			# for every forwarded onto protocol, ...
			foreach my $onto_protocol 
			    ( sort keys %$onto_protocols_ports_ref ) {

			    # every forwarded onto protocol has to have a 
			    # matching forwarded from same protocol
			    if ( ! exists $$forwarded_protocols_ports_ref{$onto_protocol} ) {
				$syntax_error = 1;
				push @directive_error_lines,
				"forwarded onto protocols include $onto_protocol which is not included in the forwarded from protocols $for $where";
                            } # end else if protocol in both onto and forwarded
			} # end foreach forwarded onto protocol

		    } # end if forwarded onto protocols/ports were parsed
		} # end else if valid forwareded onto address parsed
	    } # end if onto portion of FORWARD directive being parsed
	} # end if processinf FORWARD directive

	# handle INTIF/INT INTERFACE(S)/INTERNAL INTERFACE(S) directives
	elsif ( $fields[0] =~ /^intif$/i ||
	     ( @fields >= 2 && $fields[0] =~ /^int(ernal)?$/i ) ||
	     ( @fields >= 3 && $fields[0] =~ /^int(ernal)?$/i &&
	                       $fields[1] =~ /^interface(s)?$/i ) ) {
	    if ( $fields[0] eq "intif" ) {
		shift @fields;
	    } elsif ( $fields[1] =~ /^interface(s)?$/i ) {
		$for = "$for $fields[1]";
		$directive{directive_type} = "intif";
		shift @fields;
		shift @fields;
	    } else {
		$directive{directive_type} = "intif";
		shift @fields;
	    }

	    # has to have at least one argument
	    if (@fields < 1 ) {
		push @directive_error_lines, "not enough arguments $for $where";
		$syntax_error = 1;
	    } else {

		# parse interface name(s)
		my @interfaces = 
		    parse_interfaces( $vars_ref,
				      \@fields,
				      \@directive_error_lines,
				      \@directive_warning_lines, 
				      \$syntax_error,
				      $for,
				      $where);
		$directive{interfaces} = \@interfaces;
	    }
	}
	
	# handle LITERAL directive

	elsif ( $fields[0] eq "literal" ) {
	    shift @fields;
	    # cannot be any arguments but we will still 
	    # just issue a warning if so
	    if (@fields) {
		push @directive_warning_lines,
		"literal requires no arguments $for $where";
		@fields = ();
	    }
	    # if we're in the middle of a literal block
	    # this is an extra LITERAL, declare warning
	    if ( $in_literal_block ) {
		push @directive_warning_lines,
		"literal while already in a literal block $for $where";
	    }
	    # but if we are not in a literal block, start one
	    else {
		@literal_lines = ();
		$in_literal_block = 1;
		$literal_line_num = $line_num;
		$literal_config_line_number = $line_num;
		$literal_config_line = $config_file_line;
	    }
	    # in any event we do not want to store a directive yet
	    next;
	}

	# handle LOGGING [ none | limited | full ] directive

	elsif ( $fields[0] eq "logging" ) {
	    shift @fields;
	    # if no arguments, set to none
	    if ( ! @fields ) {
		${$$vars_ref{constants}}{logging} = "none";
	    # single argument could be none | limited | full
	    } else {
		my $argument = shift @fields;
		if ( $argument ne "none" &&
		     $argument ne "limited" &&
		     $argument ne "full" ) {
		    $syntax_error = 1;
		    push @directive_error_lines,
		    "logging can be none or limited or full";
		} elsif ( ! @fields ) {
		    ${$$vars_ref{constants}}{logging} = $argument;
	        }
	    }
	}

	# handle NOLOGGING directive

	elsif ( $fields[0] eq "nologging" ) {
	    shift @fields;
	    # if no arguments, set to none
	    if ( ! @fields ) {
		${$$vars_ref{constants}}{logging} = "none";
	    # if any arguments, syntax error
	    } else {
		$syntax_error = 1;
		push @directive_error_lines,
		"nologging cannot have any arguments";
	        shift @fields;
	    }
	}

	# handle PRIVATE PROTECTED PUBLIC UNPROTECTED directives
	#
	# syntax: 
	#
	# PRIVATE interface-name(s)
	# PROTECTED interface-name(s)
	# PUBLIC interface-name(s)
	# UNPROTECTED interface-name(s)

	elsif ( $fields[0] eq "private"     ||
	        $fields[0] eq "protected"   ||
	        $fields[0] eq "public"      ||
	        $fields[0] eq "unprotected" ) {
	    shift @fields;

	    # get rid of possible INTERFACE(S) keyword
	    if ( $fields[0] =~ /^interface(s)?$/i ) {
		$for = "$for $fields[0]";
		shift @fields;
	    }

	    # has to have at least one argument
	    if (@fields < 1 ) {
		push @directive_error_lines, "not enough arguments $for $where";
		$syntax_error = 1;
	    } else {

		# parse interface name(s)
		my @interfaces =
		    parse_interfaces( $vars_ref,
				      \@fields,
				      \@directive_error_lines,
				      \@directive_warning_lines, 
				      \$syntax_error,
				      $for,
				      $where);
                $directive{interfaces} = \@interfaces;
	    }
	}

	# unknown directive
	
	else { 
	    push @directive_error_lines, "unknown directive <$fields[0]> $where";
	    $syntax_error = 1;
	}

	# if no syntax error yet, and unparsed stuff on input line,
	# mark this directive bad and note the unparsed fields
	# in an error message
	if ( @fields && ! $syntax_error ) {
	    push @directive_error_lines,
	    "extra/unknown argument(s) <@fields> $for $where";
	    $syntax_error = 1;
	}

	# add the error and warning lines for this directive to 
	# the global error and warning lines
	push @$error_lines_ref, @directive_error_lines;
	push @$warning_lines_ref, @directive_warning_lines;

	# if there are error lines and/or warning lines, add them
	# to the directive block in case we want to use them
	# when generating the intermediate source file or the 
	# output commands file, except get rid of the location
	# string on the end of each
	if ( @directive_error_lines ) {
	    foreach my $line ( @directive_error_lines ) {
		$line =~ s/ $where//;
	    }
	    $directive{error_lines} = \@directive_error_lines;
	}
	if ( @directive_warning_lines ) {
	    foreach my $line ( @directive_warning_lines ) {
		$line =~ s/ $where//;
	    }
	    $directive{warning_lines} = \@directive_warning_lines;
	}

	# if there was a syntax error, change the directive type
	# to a "syntax_error" so that this directive can still be
	# inserted into the intermediate source file and the 
	# output commands file as a comment lines block
	if ( $syntax_error ) {
	    $directive{directive_type} = syntax_error;
	}

	# save the directive
	push @directives, \%directive;

    } # end foreach config line

    # If there are comment lines accumulated, but not
    # stored in a directive, store them now

    if ( @comment_lines ) {
	my %comment_directive = (
				 directive_type    => "comment_lines",
				 comment_lines_ref => [@comment_lines],
				 comment_lines_num => scalar @comment_lines
				 );
	push @directives, \%comment_directive;
    }

    # If there are literal block lines accumulated
    # that were not terminated, output an error,
    # but still save the accumulated literal lines

    if ( $in_literal_block ) {
	my %literal_directive = (
				 directive_type                => "literal_block",
				 processed                     => 0,
				 endliteral_config_line        => "",
				 endliteral_config_line_number => "",
				 literal_config_line           => $literal_config_line,
				 literal_config_line_number    => $literal_config_line_num,
				 literal_lines_ref             => \[@literal_lines]
				 );
	push @directives, \%literal_directive;
	push @$error_lines_ref, "unterminated literal block";
    }

#    print "$0: exiting parse_pfilter_$debug$$vars_ref{debug} verbose=${$$vars_ref{constants}}{verbose}\n"
#	if $debug;

    return ( \@directives, \@$error_lines_ref, \@$warning_lines_ref );
}

#=======================================================================

# function to parse the the ports and/or services part of a config line
#
# Passed input parameters are:
#   reference to hash of useful flags/values
#   reference to array of words/fields left to be parsed
#   reference to array of anticipated keywords
#   defined network services action either "open" or "close"
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be possibly added to
#   reference to syntax error flag
#   flag set to non-zero if allowed to return with unknown input left
#   string used to identify directive name in error lines
#   string used to identify directive location in error lines
# Returned parameters are:
#   reference to parsed services array
#   reference to parsed protocols/ports hash

sub parse_services_protocols_ports
{
    my( $vars_ref, $fields_ref, $anticipated_keywords_ref,
	$service_action, $error_lines_ref, $warning_lines_ref, 
	$syntax_error_ref, $return_ok_with_unknown_input, 
	$for, $where) = @_;
    my @current_protocols;
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    my $keywords_ref = $$vars_ref{keywords};
    my $named_services_protocols_ports_ref =
	$$vars_ref{named_services_protocols_ports};
    my $protocols_ref = $$vars_ref{protocols};
    my $protocols_was_last;
    my @parsed_services;
    my %parsed_protocols_ports;
    my $service_after_protocols;
    my $unknown_field = 0;

    if ( $debug ) {
	my $fields_string = join( ',', @$fields_ref );
	my $keywords_string = join( ',', @$anticipated_keywords_ref );
	print "$0: entering parse_services_protocols_ports keywords=($keywords_string) fields=($fields_string)\n";
    }

    # loop through arguments, stopping when there are no more arguments,
    # or when there has been a syntax error, or when the next argument is
    # a language keyword (which can't be used as a service or port name),
    # or when there is unknown input and the return_ok_with_unknown_input
    # flag has been passed true to this function
    while ( @$fields_ref && 
	    ! $$syntax_error_ref &&
	    ! ( grep( /^\L$$fields_ref[0]\E$/, @$anticipated_keywords_ref ) ) &&
	    ! ( $unknown_field && $return_ok_with_unknown_input ) ) {
#	if ( $debug ) {
#	    print "$0: in parse_services_protocols_ports parsed_services = @parsed_services\n";
#	    print "$0:         fields = @$fields_ref\n";
#	    print "$0:         current_protocols = @current_protocols\n";
#	    my $fields_string = join( ',', @$fields_ref);
#	    print "$0:         fields=($fields_string)\n";
#	    my $pwl = $protocols_was_last || "undef";
#	    print "$0:         protocols_was_last = $pwl\n";
#	    print_hash( "        ", "parsed_pp", \%parsed_protocols_ports );
#	}

	# if the current argument has embedded comma characters
	# in the middle of it, we will seperate it into more 
	# arguments, since the previous protocol allowed this,
	# but we will add a warning message

	if ( $$fields_ref[0] =~ /\,/ ) {
	    my $old_field = shift @$fields_ref;
	    push @$warning_lines_ref,
	    "Allowing services/protocols/ports seperated by commas $for $where";
	    push @$warning_lines_ref,
	    "Please seperate $old_field with spaces and/or tab characters.";
	    push @$warning_lines_ref, "This feature may be deprecated soon.";
#	    my @new_fields = split '\s*\,\s*', $old_field; 
	    my @new_fields = split '\,', $old_field; 
	    unshift @$fields_ref, @new_fields;
#	    if ( $debug ) {
#		my $string = join( ',', @$fields_ref );
#		print "$0: in parse_services_protocols_ports after splitting comma seperated old_field=$old_field fields=<$string>\n";
#	    }
	}

	# if the current argument is a protocol name, ...

	if ( $protocols_ref->{$fields_ref->[0]} ) {

	    # if the current argument is a protocol name, and there were
	    # previous protocols specified without any port names/numbers
	    # being specified, add the current argument to the current
	    # protocols list

	    if ( $protocols_was_last ) {
		my $protocol = shift @$fields_ref;
#		print "$0: in parse_services_protocols_ports additional protocol <$$fields_ref[0]>\n"
#		    if ( $debug );
		if ( grep ( /^$protocol$/, @current_protocols ) ) {
		    push @$warning_lines_ref,
		    ( "warning - protocol $protocol listed twice before ports list $where" );
		} else {
		    push @current_protocols, $protocol;
		}
	    }

	    # if the current argument is a protocol name, and the previous
	    # argument was not a protocol, start a new current protocols list

	    else {
#		print "$0: in parse_services_protocols_ports new protocol <$$fields_ref[0]>\n"
#		    if ( $debug );
		@current_protocols = shift @$fields_ref;
		$service_after_protocols = 0;
		$protocols_was_last = 1;
	    }

	    next;
	}

	# if the current argument is a service name that was
	# defined in the pfilter ruleset files, store the service

	if ( $$vars_ref{services}{$$fields_ref[0]} ) {
#	    print "$0: in parse_services_protocols_ports new ruleset defined service <$$fields_ref[0]>\n"
#		if ( $debug );
	    $service_after_protocols = 1;
	    $protocols_was_last = 0;
	    my $service_ref = $$vars_ref{services}{$$fields_ref[0]};
	    if ( exists $$service_ref{$service_action} ) {
		push @parsed_services, shift @$fields_ref;
	    } else {
		push @$error_lines_ref,
		"$0: error - ruleset defined network service $$fields_ref[0] specified with missing/inappropriate action $service_action $for $where";
		$$syntax_error_ref = 1;
	    }
	    next;
	}

	# if the argument is a numeric port number, ...

	if ( $$fields_ref[0] =~ /^\d+$/ ) {
#	    print "$0: in parse_services_protocols_ports numeric port number <$$fields_ref[0]>\n"
#		if ( $debug );
	    my $port = shift @$fields_ref;

	    # if one or more protocols have been specified,
	    # store the protocol(s)/port combination(s) as long
	    # as the port number isn't out of range

	    if ( @current_protocols ) {
		foreach my $protocol ( @current_protocols ) {
		    if ( valid_port( $vars_ref, $protocol, $port ) ) {
			protocols_ports_add_protocol_port
			    ($vars_ref, \%parsed_protocols_ports,
			     $protocol, $port);
			$service_after_protocols = 0;
			$protocols_was_last = 0;
		    } else {
			push @$error_lines_ref,
			"$0: error - invalid protocol/port <$protocol>/<$port> $for $where";
			$$syntax_error_ref = 1;
		    }
		}
	    }

	    # if no protocols have been specified, the numeric port
	    # number is a syntax error


	    else {
		push @$error_lines_ref,
		"$0: error - port number $port without any protocol(s) $for $where";
		$$syntax_error_ref = 1;
	    }

	    next;
	}

	# if the argument is a numeric port range, ...

	if ( $$fields_ref[0] =~ /^\d+\:\d+$/ ) {
#	    print "$0: in parse_services_protocols_ports numeric port range <$$fields_ref[0]>\n"
#		if ( $debug );
	    my ( $port_low, $port_high ) = split '\:', shift @$fields_ref;

	    # if the range is reversed, error
	    if ( $port_low > $port_high ) {
		push @$error_lines_ref,
		"$0: error - reversed port range $$fields_ref[0] $for $where";
		$$syntax_error_ref = 1;
	    }
		
	    # if no protocols have been specified, the numeric port
	    # range is a syntax error

	    elsif ( ! @current_protocols ) {
		push @$error_lines_ref,
		"$0: error - port number $port without any protocol(s) $for $where";
		$$syntax_error_ref = 1;
	    }

	    # if one or more protocols have been specified,
	    # store the protocol(s)/port combination(s) as long
	    # as the port numbers aren't out of range

	    else {
		foreach my $protocol ( @current_protocols ) {
		    if ( valid_port( $vars_ref, $protocol, $port_low ) &&
			 valid_port( $vars_ref, $protocol, $port_high ) ) {
			protocols_ports_add_protocol_port
			    ($vars_ref, \%parsed_protocols_ports,
			     $protocol, "$port_low:$port_high");
			$service_after_protocols = 0;
			$protocols_was_last = 0;
		    } else {
			push @$error_lines_ref,
			"$0: error - invalid protocol/port <$protocol>/<$port_low> $for $where"
			    if ( ! valid_port( $vars_ref, $protocol, $port_low ) );
			push @$error_lines_ref,
			"$0: error - invalid protocol/port <$protocol>/<$port_high> $for $where"
			    if ( ! valid_port( $vars_ref, $protocol, $port_high ) );
			$$syntax_error_ref = 1;
		    }
		}
	    }

	    next;
	}

	# if the current argument is a defined port/service name
	# from /etc/services or wherever, ...

	if ( $$named_services_protocols_ports_ref{$$fields_ref[0]} ) {
#	    print "$0: in parse_services_protocols_ports /etc/services defined service <$$fields_ref[0]>\n"
#		if ( $debug );
	    my $name = shift @$fields_ref;
	    my $service_protocols_ports_ref =
		$$named_services_protocols_ports_ref{$name};

	    # if one or more protocols have been specified, and at
	    # least one of them is also in the protocols defined for this
	    # service, store the protocol(s)/ports from the service that
	    # intersect the already specified protocols

	    my @intersected_protocols = ();
	    foreach my $protocol ( @current_protocols ) {
		if ( exists $$service_protocols_ports_ref{$protocol} ) {
		    push @intersected_protocols, $protocol;
		}
	    }
	    if ( @intersected_protocols ) {
#		print "$0:         intersected protocols <@intersected_protocols>\n"
#		    if ( $debug );
		foreach my $protocol ( @intersected_protocols ) {
		    my $ports_ref = $$service_protocols_ports_ref{$protocol};
		    foreach my $port ( keys %$ports_ref ) {
			protocols_ports_add_protocol_port
			    ($vars_ref, \%parsed_protocols_ports,
			     $protocol, $port);
		    }
		}
	    }

	    # if no protocols have been specified yet, add all the
	    # protocol/port combinations defined by this service

	    else {
		protocols_ports_add_protocols_ports
		    ($vars_ref, \%parsed_protocols_ports,
		     $service_protocols_ports_ref);
	    }

	    $protocols_was_last = 0;
	    next;
	}

	# un-recognized protocol/port
	if ( $return_ok_with_unknown_input ) {
	    $unknown_field = 1;
	} else {
	    push @$error_lines_ref,
	    "$0: error - unknown service/port <$$fields_ref[0]> before protocols $for $where";
	    $$syntax_error_ref = 1;
	}

    } # end of loop parsing fields
    if ( $debug ) {
	my $fields_string = join( ',', @$fields_ref);
	my $keywords_string = join( ',', @$anticipated_keywords_ref );
	print "$0: exiting  parse_services_protocols_ports keywords=($keywords_string) fields=($fields_string)\n";
	print_hash( "", "returning parsed_protocols_ports",
		    \%parsed_protocols_ports )
	    if keys %parsed_protocols_ports;
	if ( @parsed_services ) {
	    my $string = join( ', ', @parsed_services );
	    print "$0: in parse_services_protocols_ports parsed_services = ($string)\n";
	}
    }

    return ( \@parsed_services, \%parsed_protocols_ports );
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"

1;



